

class HobokenException(Exception):
    """Base class for all Hoboken exceptions."""
    pass


class InvalidMatchTypeException(HobokenException):
    """
    Invalid matcher type given for a function.  Valid types are:
      - A string, in either splat ("/blah/*/blah") or named-parameter
        ("/blah/:param") format.
      - A regex object (e.g. re.compile("/blah/(.+)"))
      - A custom object that provides a match() callable.
    """
    pass


class RouteExistsException(HobokenException):
    """
    Exception raised when trying to set multiple route decorators on a function.
    """
    pass


class HobokenUserException(HobokenException):
    """Base class for all user-raise-able exceptions."""
    pass


class ContinueRoutingException(HobokenUserException):
    """
    This exception signals that Hoboken should continue routing the current
    request.  The current route is ignored, and the next one is tried.
    """
    pass


class HaltRoutingException(HobokenUserException):
    """
    This exception type is the base class for all exceptions that take
    arguments to be set on the Response() object.  It simply stores them for
    use by the application.
    """
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        super(HaltRoutingException, self).__init__()




from . import main, suite

# This simply runs the main test suite, so we can call this test suite as a
# module: e.g. "python -m shift.tests"
# NOTE: We need the suite import from above for this to work.
main()




# -*- coding: utf-8 -*-

from . import HobokenTestCase
from .. import HobokenApplication, condition
from ..application import Route, halt, pass_route
from ..matchers import RegexMatcher
from ..exceptions import *
import re
import sys
import unittest
from webob import Request
import mock


class TestHasHTTPMethods(HobokenTestCase):
    def test_has_all_methods(self):
        for x in self.app.SUPPORTED_METHODS:
            assert hasattr(self.app, x.lower())


class TestWorksWithConditions(HobokenTestCase):
    def after_setup(self):
        self.calls = []

        def cond_above(req):
            self.calls.append("above")
            return True

        def cond_below(req):
            self.calls.append("below")
            return True

        @condition(cond_above)
        @self.app.get('/')
        @condition(cond_below)
        def route_func():
            self.calls.append("body")
            return 'success'

    def test_condtions_order(self):
        self.assert_body_is("success")
        self.assert_equal(self.calls, ["below", "above", "body"])


class TestConditionCanAbortRequest(HobokenTestCase):
    def after_setup(self):
        def no_foo_in_path(req):
            return req.path.find('foo') == -1

        @condition(no_foo_in_path)
        @self.app.get('/:param')
        def bar(param=None):
            return param

    def test_should_work(self):
        self.assert_body_is("works", path='/works')

    def test_should_not_work(self):
        self.assert_not_found(path='/foobreaks')


class TestSubapps(HobokenTestCase):
    def after_setup(self):
        subapp = HobokenApplication("subapp")
        self.app.set_subapp(subapp)

        @subapp.get("/subapp")
        def subapp_func():
            return "subapp"

        @self.app.get("/app")
        def app_func():
            return "app"

    def test_app_call_works(self):
        self.assert_body_is("app", path='/app')

    def test_subapp_delegation_works(self):
        self.assert_body_is("subapp", path='/subapp')

    def test_neither(self):
        self.assert_not_found(path='/neither')


class TestHandlesExceptions(HobokenTestCase):
    def after_setup(self):
        @self.app.get("/errors")
        def errorme():
            raise Exception("foobar bloo blah")

    def test_exception_handling(self):
        code, body = self.call_app(path='/errors')
        self.assert_equal(code, 500)


class TestBodyReturnValues(HobokenTestCase):
    def after_setup(self):
        @self.app.get("/bytes")
        def bytes():
            return b'byte string'

        @self.app.get("/string")
        def string():
            # føø
            return b'f\xc3\xb8\xc3\xb8'.decode('utf-8')

    def test_bytes(self):
        req = Request.blank('/bytes')
        resp = req.get_response(self.app)
        self.assert_equal(resp.body, b'byte string')

    def test_bytes(self):
        req = Request.blank('/string')
        resp = req.get_response(self.app)
        self.assert_equal(resp.text, b'f\xc3\xb8\xc3\xb8'.decode('utf-8'))


class TestHaltHelper(HobokenTestCase):
    def after_setup(self):
        self.halt_code = None
        self.halt_body = None

        @self.app.before("/before/halt")
        def before_halt_func():
            halt(code=self.halt_code, text=self.halt_body)

        @self.app.get("/halts")
        def halts():
            halt(code=self.halt_code, text=self.halt_body)
            return 'bad'

        self.app.debug = True

    def assert_halts_with(self, code, body, *args, **kwargs):
        """Helper function to set the halt value and assert"""
        self.halt_code = code

        # The 'text' attribute of a webob Request only supports unicode
        # strings on Python 2.X, so we need to make this unicode.
        if sys.version_info[0] < 3:
            self.halt_body = unicode(body)
        else:
            self.halt_body = body

        self.assert_body_is(body, *args, **kwargs)

    def test_before_can_halt(self):
        self.assert_halts_with(200, 'foobar', path='/before/halt')

    def test_body_can_halt(self):
        self.assert_halts_with(200, 'good', path='/halts')


class TestPassHelper(HobokenTestCase):
    def after_setup(self):
        @self.app.get("/aroute/*")
        def pass_one(splat):
            pass_route()
            return 'bad'

        @self.app.get("/aroute/*")
        def real_route(splat):
            return 'good'

        @self.app.before("/pass/before")
        def pass_before():
            pass_route()
            self.app.response.text = 'bad'

        @self.app.before("/pass/*")
        def before_pass_all(splat):
            self.app.response.text += 'good'

        @self.app.get("/pass/*")
        def pass_before_route(splat):
            self.app.response.text += 'foo'

        self.app.debug = True

    def test_pass_route(self):
        self.assert_body_is('good', path='/aroute/')

    def test_pass_before(self):
        # Passing in filter will simply jump to the next filter.  It has no
        # effect on the actual body routes themselves.
        self.assert_body_is('goodfoo', path='/pass/before')
        self.assert_body_is('goodfoo', path='/pass/other')


class TestRedirectHelper(HobokenTestCase):
    def after_setup(self):
        self.redirect_code = 0

        @self.app.post("/upload")
        def upload():
            # Upload stuff here.
            self.app.redirect("/uploaded")

        @self.app.get("/uploaded")
        def uploaded():
            return 'uploaded successfully'

        @self.app.get("/redirect")
        def redirect_func():
            self.app.redirect('/foo', status_code=self.redirect_code)

        self.app.debug = True

    def test_redirect(self):
        req = Request.blank("/upload", method='POST')
        resp = req.get_response(self.app)

        self.assert_equal(resp.status_int, 302)
        self.assert_true(resp.location.endswith('/uploaded'))

    def test_redirect_code(self):
        for code in [301, 302, 303]:
            self.redirect_code = code

            req = Request.blank("/redirect")
            resp = req.get_response(self.app)

            self.assert_equal(resp.status_int, code)
            self.assert_true(resp.location.endswith('/foo'))

    def test_redirect_with_non_get(self):
        req = Request.blank("/upload", method='POST')
        req.http_version = "HTTP/1.1"
        resp = req.get_response(self.app)

        self.assert_equal(resp.status_int, 303)
        self.assert_true(resp.location.endswith('/uploaded'))


class TestRoute(HobokenTestCase):
    def test_route_uppercases_method(self):
        m = Route(None, None)
        m.method = 'get'
        self.assert_equal(m.method, 'GET')


class TestMatcherTypes(HobokenTestCase):
    def test_will_handle_regex(self):
        r = re.compile("(.*?)")
        @self.app.get(r)
        def regex_get():
            return b'body'

        route = self.app.find_route(regex_get)
        self.assert_true(isinstance(route.matcher, RegexMatcher))

    def test_will_handle_regex_named_captures(self):
        r = re.compile("/(.*?)foo(?P<name>.*?)bar")

        @self.app.get(r)
        def regex_get_params(arg, name=None):
            self.assert_equal(arg, 'ONE')
            self.assert_equal(name, 'TWO')
            return b'param'

        r = Request.blank("/ONEfooTWObar")
        resp = r.get_response(self.app)
        self.assert_equal(resp.status_code, 200)
        self.assert_equal(resp.body, b'param')

    def test_will_handle_regex_named_captures_2(self):
        r = re.compile("/(?P<first>.*?)foo(?P<second>.*?)bar")

        @self.app.get(r)
        def regex_get_params(first=None, second=None):
            self.assert_equal(first, 'ONE')
            self.assert_equal(second, 'TWO')
            return b'param'

        r = Request.blank("/ONEfooTWObar")
        resp = r.get_response(self.app)
        self.assert_equal(resp.status_code, 200)
        self.assert_equal(resp.body, b'param')

    def test_will_handle_regex_named_captures_3(self):
        r = re.compile("/(.*?)foo(.*?)bar")

        @self.app.get(r)
        def regex_get_params(arg1, arg2):
            self.assert_equal(arg1, 'ONE')
            self.assert_equal(arg2, 'TWO')
            return b'param'

        r = Request.blank("/ONEfooTWObar")
        resp = r.get_response(self.app)
        self.assert_equal(resp.status_code, 200)
        self.assert_equal(resp.body, b'param')

    def test_will_handle_regex_named_captures_4(self):
        r = re.compile("/(?P<first>.*?)foo(.*?)bar")

        @self.app.get(r)
        def regex_get_params(arg, first=None):
            self.assert_equal(first, 'ONE')
            self.assert_equal(arg, 'TWO')
            return b'param'

        r = Request.blank("/ONEfooTWObar")
        resp = r.get_response(self.app)
        self.assert_equal(resp.status_code, 200)
        self.assert_equal(resp.body, b'param')

    def test_will_handle_custom_matcher(self):
        m = mock.MagicMock()
        m.match.return_value = (True, ['arg'], {'val': 'kwarg'})

        @self.app.get(m)
        def custom_get(arg, val=None):
            self.assert_equal(arg, 'arg')
            self.assert_equal(val, 'kwarg')
            return b'body'

        r = Request.blank("/")
        resp = r.get_response(self.app)
        self.assert_equal(resp.status_code, 200)
        self.assert_equal(m.match.call_count, 1)


class TestMiscellaneousMethods(HobokenTestCase):
    def test_add_route_will_throw(self):
        with self.assert_raises(HobokenException):
            self.app.add_route('bad', None, None)

    def test_invalid_matcher_type(self):
        with self.assert_raises(InvalidMatchTypeException):
            @self.app.get(123)
            def bad():
                pass

    def test_find_route_will_return_none_on_failure(self):
        def not_exist():
            pass

        res = self.app.find_route(not_exist)
        self.assert_true(res is None)

    def test_url_for_will_return_none_on_failure(self):
        def not_a_route():
            pass

        url = self.app.url_for(not_a_route)
        self.assert_true(url is None)

    def test_only_one_route_per_function(self):
        with self.assert_raises(RouteExistsException):
            @self.app.get("/one")
            @self.app.get("/two")
            def route_func():
                pass

    def test_invalid_method(self):
        @self.app.get("/")
        def index():
            return b'foo'

        r = Request.blank("/")
        r.method = 'OTHER'
        resp = r.get_response(self.app)

        self.assert_equal(resp.status_code, 405)


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestHasHTTPMethods))
    suite.addTest(unittest.makeSuite(TestWorksWithConditions))
    suite.addTest(unittest.makeSuite(TestConditionCanAbortRequest))
    suite.addTest(unittest.makeSuite(TestSubapps))
    suite.addTest(unittest.makeSuite(TestHandlesExceptions))
    suite.addTest(unittest.makeSuite(TestBodyReturnValues))
    suite.addTest(unittest.makeSuite(TestHaltHelper))
    suite.addTest(unittest.makeSuite(TestPassHelper))
    suite.addTest(unittest.makeSuite(TestRedirectHelper))
    suite.addTest(unittest.makeSuite(TestRoute))
    suite.addTest(unittest.makeSuite(TestMatcherTypes))
    suite.addTest(unittest.makeSuite(TestMiscellaneousMethods))

    return suite




from . import HobokenTestCase, skip
import unittest

class TestFilters(HobokenTestCase):
    def after_setup(self):
        self.calls = []

        @self.app.before("/before/*")
        def before_filter(splat):
            self.calls.append("before")

        @self.app.get("/*")
        def route_func(splat):
            self.calls.append("body")
            return splat

        @self.app.after("/after/*")
        def after_filter(splat):
            self.calls.append("after")

        @self.app.before("/both/*")
        def both_before_filter(splat):
            self.calls.append("before")

        @self.app.after("/both/*")
        def both_after_filter(splat):
            self.calls.append("after")

        @self.app.after("/both_at_once/*")
        @self.app.before("/both_at_once/*")
        def both_at_once_filter(splat):
            self.calls.append("both_at_once")

    def test_neither(self):
        self.assert_body_is("neither", path="/neither")
        self.assert_equal(self.calls, ["body"])

    def test_before(self):
        self.assert_body_is("before/stuff", path="/before/stuff")
        self.assert_equal(self.calls, ["before", "body"])

    def test_after(self):
        self.assert_body_is("after/stuff", path="/after/stuff")
        self.assert_equal(self.calls, ["body", "after"])

    def test_both(self):
        self.assert_body_is("both/stuff", path="/both/stuff")
        self.assert_equal(self.calls, ["before", "body", "after"])

    def test_both_at_once(self):
        self.assert_body_is("both_at_once/stuff", path="/both_at_once/stuff")
        self.assert_equal(self.calls, ["both_at_once", "body", "both_at_once"])


class TestFilterCanModifyRoute(HobokenTestCase):
    def after_setup(self):
        @self.app.before("/notmatched")
        def modify_route():
            self.app.request.environ['PATH_INFO'] = "/matched"

        @self.app.get("/matched")
        def route_func():
            return 'success'

    def test_matched_route(self):
        self.assert_body_is("success", path="/matched")

    def test_modified_route(self):
        self.assert_body_is("success", path="/notmatched")


class TestFilterParams(HobokenTestCase):
    def after_setup(self):
        @self.app.before("/before/:param1/:param2/*")
        def before_func(splat, param1=None, param2=None):
            self.val = (param1 + '\n' +
                        param2 + '\n' +
                        splat)

        @self.app.after("/after/:foo/*/:bar")
        def after_func(splat, foo=None, bar=None):
            self.val = (foo + '\n' +
                        splat + '\n' +
                        bar)

        @self.app.get("/*")
        def catchall(splat):
            pass

        self.app.debug = True

    def test_before_params(self):
        self.call_app(path='/before/one/two/params')
        self.assert_equal(self.val, 'one\ntwo\nparams')

    def test_after_params(self):
        self.call_app(path='/after/abcd/morestuff/defg')
        self.assert_equal(self.val, 'abcd\nmorestuff\ndefg')


class TestCatchallFilters(HobokenTestCase):
    def after_setup(self):
        self.calls = []

        @self.app.before()
        def before_all():
            self.calls.append('before')

        @self.app.after()
        def after_all():
            self.calls.append('after')

        @self.app.get("/")
        def index():
            self.calls.append('body')
            return b'body'

    def test_catchall_filters(self):
        self.call_app(path='/')
        self.assert_equal(self.calls, ['before', 'body', 'after'])


def suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestFilters))
    suite.addTest(unittest.makeSuite(TestFilterCanModifyRoute))
    suite.addTest(unittest.makeSuite(TestFilterParams))
    suite.addTest(unittest.makeSuite(TestCatchallFilters))

    return suite




from __future__ import with_statement, absolute_import, print_function

from .application import halt
import sys
import time
import datetime
import webob


class HobokenCachingMixin(object):
    """
    This class defines a mixin that one can combine with the base
    HobokenApplication to add some handy helpers for dealing with caching.
    """
    def __init__(self):
        raise NotImplementedError("Hoboken mixins cannot be instantiated!")

    def check_last_modified(self, date):
        """
        This function will check if one of a given request's last modified headers
        are set, and, if so, will check it against the date provided.  If the
        request specifies an equivalent or newer resource, this function will call
        halt() to abort the current request with a 304 Not Modified status.
        """
        if date is None:
            return

        # Python's time functions are stupid.  We do everything with unix times.
        timestamp = time.mktime(date.timetuple())
        self.response.last_modified = timestamp

        # We don't do anything if there's an ETag.
        if self.request.if_none_match is not webob.etag.NoETag:
            return

        if self.response.status_int == 200 and self.request.if_modified_since is not None:
            time_val = time.mktime(self.request.if_modified_since.timetuple())
            if time_val >= timestamp:
                halt(status_code=304)

        if ((self.response.is_success or self.response.status_int == 412) and
             self.request.if_unmodified_since is not None):
            time_val = time.mktime(self.request.if_unmodified_since.timetuple())
            if time_val < timestamp:
                halt(status_code=412)

    def check_etag(self, etag, new_resource=False, weak=False):
        """
        As per check_if_modified(), except checks the ETag header instead.
        """
        self.response.etag = (etag, not weak)

        # We assume the request is a new resource if it is a POST.
        new_resource = new_resource or self.request.method == "POST"

        # An etag will match a 'If-*-Match' header in two cases:
        #  - If it's not a new resource, and the header specifies 'anything' (i.e. '*')
        #  - Otherwise, if it's an exact match.
        def etag_matches(value):
            if value is webob.etag.AnyETag:
                return not new_resource
            return self.response.etag in value

        if self.response.is_success or self.response.status_int == 304:
            if self.request.if_none_match is not webob.etag.NoETag and etag_matches(self.request.if_none_match):
                if self.request.is_safe:
                    halt(status_code=304)
                else:
                    halt(status_code=412)
            elif self.request.if_match is not webob.etag.AnyETag and not etag_matches(self.request.if_match):
                halt(status_code=412)

    def set_cache_control(self, **kwargs):
        for key, val in kwargs.items():
            setattr(self.response.cache_control, key, val)

    def set_expires(self, amount, **kwargs):
        if isinstance(amount, int):
            max_age = amount
            amount = datetime.datetime.now() + datetime.timedelta(seconds=amount)
        else:
            now = datetime.datetime.now()
            if now >= amount:
                # Do nothing, this expired already.
                max_age = 0
            else:
                max_age = (amount - now).seconds

        kwargs['max_age'] = max_age
        self.set_cache_control(**kwargs)
        self.response.expires = amount


class HobokenRedirectMixin(object):
    def __init__(self):
        raise NotImplementedError("Hoboken mixins cannot be instantiated!")

    def redirect_back(self, *args, **kwargs):
        """
        This is a helper function to redirect 'back' - i.e. to whichever page
        referred to this one.
        TODO: maybe emit HTML to redirect back if no referrer?
        """
        if self.request.referer is not None and self.request.referer != '':
            self.redirect(location=self.request.referer, *args, **kwargs)
        # TODO: do we want to do this?
        # elif 'text/html' in self.request.accept:
        #     body_text = """
        #     <html><body onload="history.go(-1);">Please go back one page</body></html>
        #     """

        #     halt(status_code=200, text=body_text)
        else:
            return False

    def redirect_to(self, func, status_code=None, *args, **kwargs):
        """
        This is a helper function to redirect to another route.
        """
        location = self.url_for(func, *args, **kwargs)
        self.redirect(location, status_code=status_code)


class HobokenRenderMixin(object):
    def __init__(self):
        raise NotImplementedError("Hoboken mixins cannot be instantiated!")

    # Mixin __init__() calls are not called, so we initialize shift here.
    # import shift
    # _shift = shift.Shift()

    # TODO: might want to make Shift a hard dependency, so we can configure it properly

    def render(self, file, context=None):
        # self._shift
        pass




# Future-proofing
from __future__ import with_statement, absolute_import, print_function

import webob

class Request(webob.BaseRequest):
    """
    Hoboken's request objects.  Uses WebOb's request object
    """
    @property
    def is_safe(self):
        """
        Returns True if the request is considered "safe" - i.e. if the request
        should be treated as not modifying any state.
        """
        return self.method in ['GET', 'HEAD', 'OPTIONS', 'TRACE']

    @property
    def is_idempotent(self):
        """
        Returns True if the request is considered idempotent - i.e. if two
        successive identical requests should result in the same result.
        """
        return self.is_safe or self.method in ['PUT', 'DELETE']


class Response(webob.Response):
    """
    Hoboken's request objects.  Uses WebOb's request object.
    """
    @property
    def is_informational(self):
        return 100 <= self.status_int <= 199

    @property
    def is_success(self):
        return 200 <= self.status_int <= 299

    @property
    def is_redirect(self):
        return 300 <= self.status_int <= 399

    @property
    def is_client_error(self):
        return 400 <= self.status_int <= 499

    @property
    def is_server_error(self):
        return 500 <= self.status_int <= 599

    @property
    def is_not_found(self):
        return self.status_int == 404





import re
import sys

__all__ = [
    "iscallable", "BaseStringType", "StringIO", "RegexType", "get_func_attr",
    "set_func_attr"
]

# Determine Python version.
python_version = sys.version_info[:3]
is_python3 = python_version >= (3, 0, 0)

if is_python3:
    import collections

    # callable() replacement
    def iscallable(f):
        return isinstance(f, collections.Callable)

    # basestring
    BaseStringType = str

    # StringIO
    from io import StringIO

    def get_func_attr(func, attr, default=None, delete=False):
        if delete:
            return func.__dict__.pop(attr, default)
        else:
            return func.__dict__.get(attr, default)

    def set_func_attr(func, attr, value):
        func.__dict__[attr] = value

else:
    def iscallable(f):
        return callable(f)

    # basestring
    BaseStringType = basestring

    # StringIO
    try:
        from cStringIO import StringIO
    except ImportError:
        from StringIO import StringIO

    def get_func_attr(func, attr, default=None, delete=False):
        if delete:
            return func.func_dict.pop(attr, default)
        else:
            return func.func_dict.get(attr, default)

    def set_func_attr(func, attr, value):
        func.func_dict[attr] = value


# Cross-Python stuff.
RegexType = type(re.compile(""))
RegexMatchType = type(re.compile(".*").match("asdf"))



from __future__ import with_statement

import os
import sys
import types
import unittest
import functools


def ensure_in_path(path):
    """
    Ensure that a given path is in the sys.path array
    """
    if not os.path.isdir(path):
        raise RuntimeError('Tried to add nonexisting path')

    def _samefile(x, y):
        try:
            return os.path.samefile(x, y)
        except (IOError, OSError):
            return False
        except AttributeError:
            # Probably on Windows.
            path1 = os.path.abspath(x).lower()
            path2 = os.path.abspath(y).lower()
            return path1 == path2

    # Remove existing copies of it.
    for pth in sys.path:
        if _samefile(pth, path):
            sys.path.remove(pth)

    # Add it at the beginning.
    sys.path.insert(0, path)


# This will pull in the skip() function from the unittest module, or, if it
# doesn't exist, will make a decorator that simply returns a dummy class or
# function, as appropriate.
if hasattr(unittest, 'skip'):
    skip = unittest.skip
else:
    def skip(reason):
        def decorator(obj):
            if isinstance(obj, object):
                class Nothing(object):
                    pass

                return Nothing
            else:
                def internal_function(*args, **kwargs):
                    return None
                return internal_function

        return decorator

if hasattr(unittest, 'skipIf'):
    skip_if = unittest.skipIf
else:
    def skip_if(condition, reason):
        if condition:
            return skip(reason)
        else:
            def nop(obj):
                return obj
            return nop


def is_pypy():
    return hasattr(sys, 'pypy_version_info')

def is_python3():
    return sys.version_info[0] >= 3


class _ExceptionCatcher(object):
    """
    This is a context manager that asserts that a particular exception was raised
    during it.  This was borrowed from Mitsuhiko's flask testing code.
    """
    def __init__(self, test_case, exc_type):
        self.test_case = test_case
        self.exc_type = exc_type

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        exception_name = self.exc_type.__name__
        if exc_type is None:
            self.test_case.fail('Expected exception of type %r' % exception_name)
        elif not issubclass(exc_type, self.exc_type):
            raise exc_type(exc_value)
        return True


class BaseTestCase(unittest.TestCase):
    """
    This TestCase subclass adds some useful aliases for the camelCased names
    in the standard library (like nose).
    """

    def assert_equal(self, x, y):
        return self.assertEqual(x, y)

    def assert_not_equal(self, x, y):
        return self.assertNotEqual(x, y)

    def assert_true(self, x):
        return self.assertTrue(x)

    def assert_false(self, x):
        return self.assertFalse(x)

    def assert_in(self, x, y):
        assert x in y, "{0!r} is not in {1!r}".format(x, y)

    def assert_raises(self, exception, callable=None, *args, **kwargs):
        catcher = _ExceptionCatcher(self, exception)
        if callable is None:
            return catcher
        with catcher:
            callable(*args, **kwargs)

    def assert_is_instance(self, obj, type):
        assert isinstance(obj, type), "{0!r} is not an instance of type {1!r}".format(obj, type)

    def setup(self):
        """This is a non-camelCase hook that is identical to setUp"""
        pass

    def teardown(self):
        """This is a non-camelCase hook that is identical to tearDown"""
        pass

    def setUp(self):
        self.setup()

    def tearDown(self):
        unittest.TestCase.tearDown(self)
        self.teardown()


def parameters(params_list, name_func=None):
    """
    This function decorates a function, marking it as a function in a class
    to be parametrized.  This is then used by the parametrize() class decorator
    and associated metaclass, below, to dynamically replace copies of the
    associated methods with parametrized versions.
    """
    def internal_decorator(func):
        if is_python3():
            func.__dict__['params'] = params_list
            func.__dict__['name_func'] = name_func
        else:
            func.func_dict['params'] = params_list
            func.func_dict['name_func'] = name_func
        return func

    return internal_decorator


class ParametrizingMetaclass(type):
    def __new__(klass, name, bases, attrs):
        new_attrs = attrs.copy()
        for attr_name, attr in attrs.items():
            # We only care about functions
            if not isinstance(attr, types.FunctionType):
                continue

            if is_python3():
                func_dict = attr.__dict__
            else:
                func_dict = attr.func_dict

            # If the function doesn't have a 'params' attribute,
            # we ignore it.
            if 'params' not in func_dict:
                continue
            else:
                params = func_dict['params']

            # Either use the name function...
            name_func = func_dict['name_func']

            # ... or default to one that just returns the function name plus
            # the number of the parameter.
            if name_func is None:
                name_func = lambda i, p: attr_name + str(i)

            for i, param in enumerate(params):
                # Get the name for this parameter.
                new_name = name_func(i, param)

                # This is a pseudo-decorator that, given a function and a
                # parameter, will return a function that will call the
                # function with the given parameter.  The only difference is
                # that it will set the __name__ parameter of the function to
                # the new_name value from the scope, above.
                def create_method(func, call_param):
                    old_method = func

                    # functools.wraps() here ensures we copy the various
                    # special attributes like __doc__, and the function
                    # dict, over.
                    @functools.wraps(func)
                    def new_method(self):
                        return old_method(self, call_param)

                    # We manually override the name, since the wraps() method
                    # might clobber that.
                    new_method.__name__ = new_name

                    if is_python3():
                        new_dict = new_method.__dict__
                    else:
                        new_dict = new_method.func_dict

                    # Also, we remove the two attributes we've added from the
                    # function's dict (but don't error if they're not there).
                    new_dict.pop("params", None)
                    new_dict.pop("name_func", None)
                    return new_method

                # We create our parametrized method, and then assign it to our
                # new arguments dict.
                parametrized_method = create_method(attr, param)
                new_attrs[new_name] = parametrized_method

            # Remove the old attribute from the new dict.
            del new_attrs[attr_name]

        # We create the class as normal, except we use our new attributes.
        return type.__new__(klass, name, bases, new_attrs)


def parametrize(klass):
    """
    This is a class decorator that simply creates the class as normal, except
    that it uses the ParametrizingMetaclass in a Python 2/3 agnostic manner.
    """
    return ParametrizingMetaclass(klass.__name__, klass.__bases__, klass.__dict__)


